// Our UI is just a tray icon, which serves only to let us know the app is
// running and give us some way to close it.
// This code adapted from code by "wharfinger" @ http://everything2.com/title/Win32+system+tray+icon

#include <windows.h>
#include "fxc.h"
#include "fxcHook.h"
#include "fxcMessage.h"
#include "resources.h"

enum {
   ID_TRAYICON         = 1,
   APPWM_TRAYICON      = WM_APP,
   APPWM_NOP           = WM_APP + 1,
   ID_ABOUT            = 2000,
   ID_EXIT,
};

static BOOL onCommand(HWND hwnd, WORD commandID, HWND hwndControl)
{
   static LPCSTR HELP_ABOUT =
      "Monitors Cubase and forces \"Always on Top\" windows to ACTUALLY be on top.\n\n"
      "erictetz@gmail.com";

   switch (commandID)
   {
   case ID_ABOUT:
      MessageBox(hwnd, HELP_ABOUT, APP_TITLE, MB_ICONINFORMATION | MB_OK | MB_TOPMOST);
      return 0;

   case ID_EXIT:
      PostMessage(hwnd, WM_CLOSE, 0, 0);
      return 0;
   }

   return 0;
}

static HICON loadSmallIcon()
{
   return (HICON) LoadImage(GetModuleHandle(NULL), MAKEINTRESOURCE(TRAY_ICON_ID), IMAGE_ICON, 16, 16, 0);
}

static void addTrayIcon(HWND hwnd, UINT trayIconID, UINT callbackMsg, LPCSTR toolTip)
{
   NOTIFYICONDATA  nid;

   memset(&nid, 0, sizeof(nid));

   nid.cbSize              = sizeof(nid);
   nid.hWnd                = hwnd;
   nid.uID                 = trayIconID;
   nid.uFlags              = NIF_ICON | NIF_MESSAGE | NIF_TIP;
   nid.uCallbackMessage    = callbackMsg;
   nid.hIcon               = loadSmallIcon();

   strcpy(nid.szTip, toolTip);

   Shell_NotifyIcon(NIM_ADD, &nid);
}

static void modifyTrayIcon(HWND hwnd, UINT trayIconID, LPSTR toolTip)
{
   NOTIFYICONDATA  nid;

   memset(&nid, 0, sizeof(nid));

   nid.cbSize  = sizeof(nid);
   nid.hWnd    = hwnd;
   nid.uID     = trayIconID;
   nid.hIcon   = loadSmallIcon();
   nid.uFlags  |= NIF_ICON;

   if (toolTip)
   {
      strcpy(nid.szTip, toolTip);
      nid.uFlags  |= NIF_TIP;
      Shell_NotifyIcon(NIM_MODIFY, &nid);
   }
}

static void removeTrayIcon(HWND hwnd, UINT trayIconID)
{
   NOTIFYICONDATA  nid;

   memset(&nid, 0, sizeof(nid));

   nid.cbSize  = sizeof(nid);
   nid.hWnd    = hwnd;
   nid.uID     = trayIconID;

   Shell_NotifyIcon(NIM_DELETE, &nid);
}

static void showPopupMenu(HWND hwnd)
{
   HMENU hpopup = CreatePopupMenu();

   int menuItem = 0;
   InsertMenu(hpopup, menuItem++, MF_BYPOSITION | MF_STRING, ID_ABOUT, "About...");
   InsertMenu(hpopup, menuItem++, MF_BYPOSITION | MF_STRING, ID_EXIT, "Exit");

   POINT point;
   GetCursorPos(&point);

   SetMenuDefaultItem(hpopup, ID_ABOUT, FALSE);

   SetFocus(hwnd);

   SendMessage(hwnd, WM_INITMENUPOPUP, (WPARAM)hpopup, 0);

   int popupFlags = TPM_LEFTALIGN | TPM_RIGHTBUTTON | TPM_RETURNCMD | TPM_NONOTIFY;
   WORD cmd = TrackPopupMenu(hpopup, popupFlags, point.x, point.y, 0, hwnd, NULL);

   SendMessage(hwnd, WM_COMMAND, cmd, 0);

   DestroyMenu(hpopup);
}


static void onTrayRightButtonClick(HWND hwnd)
{
   // This SetForegroundWindow() and PostMessage(, APPWM_NOP, ,) are 
   // recommended in some MSDN sample code; apparently there's a bug in most if 
   // not all versions of Windows: Tray icon menus don't vanish properly when 
   // cancelled unless you provide special coddling.  
   //
   // In MSDN, see: "PRB: Menus for Notification Icons Don't Work Correctly", 
   // Q135788:
   //
   // mk:@ivt:kb/Source/win32sdk/q135788.htm
   //
   // Example code:
   //
   // mk:@ivt:pdref/good/code/graphics/gdi/setdisp/c3447_7lbt.htm
   //
   // Both of these pseudo-URL's are from the July 1998 MSDN.  Those geniuses 
   // have since completely re-broken MSDN at least once, so the pseudo-URL's 
   // are useless with more recent MSDN's.
   //
   // In the April 2000 MSDN, you can search titles for "Menus for 
   // Notification Icons"; "Don't" in the title has been changed to "Do Not", 
   // and searching for either complete title doesn't work anyway.  Good old 
   // MSDN.
   

   SetForegroundWindow(hwnd);
   showPopupMenu(hwnd);
   PostMessage(hwnd, APPWM_NOP, 0, 0);
}


static LRESULT CALLBACK WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
   switch (uMsg) {
   case WM_CREATE:
      addTrayIcon(hwnd, ID_TRAYICON, APPWM_TRAYICON, APP_TITLE);
      return 0;

   case APPWM_NOP:
      return 0;

   case APPWM_TRAYICON:
      SetForegroundWindow(hwnd);

      switch (lParam) {
      case WM_RBUTTONUP:
         onTrayRightButtonClick(hwnd);
         return 0;

      case WM_LBUTTONDBLCLK:
         SendMessage(hwnd, WM_COMMAND, ID_ABOUT, 0);
         return 0;
      }
      return 0;

   case WM_COMMAND:
      return onCommand(hwnd, LOWORD(wParam), (HWND)lParam);

   case WM_CLOSE:
      removeTrayIcon(hwnd, ID_TRAYICON);
      PostQuitMessage(0);
      return DefWindowProc(hwnd, uMsg, wParam, lParam);

   default:
      return DefWindowProc(hwnd, uMsg, wParam, lParam);
   }
}


int startUI(HINSTANCE hinstance)
{
   static LPCSTR THIS_CLASSNAME = "fxcWindow";

   // bail if we already have a instance running
   HWND hprevInstance = FindWindow(THIS_CLASSNAME, APP_TITLE);
   if (hprevInstance != NULL)
   {
      return 0;
   }

   // create hidden window to receive message
   WNDCLASSEX wclx;
   memset(&wclx, 0, sizeof(wclx));
   wclx.cbSize         = sizeof(wclx);
   wclx.lpfnWndProc    = &WindowProc;
   wclx.hInstance      = hinstance;
   wclx.hCursor        = LoadCursor(NULL, IDC_ARROW);
   wclx.hbrBackground  = (HBRUSH)(COLOR_BTNFACE + 1); //  COLOR_* + 1 is
   wclx.lpszClassName  = THIS_CLASSNAME;
   RegisterClassEx(&wclx);
   HWND hwnd = CreateWindow(THIS_CLASSNAME, APP_TITLE, 0, 0, 0, 100, 100, NULL, NULL, hinstance, NULL);
   if (hwnd == NULL) {
      errorMessage("Can't create our main window. We can't go on, sorry!");
      return 1;
   }

   // message loop
   MSG msg;
   while (GetMessage(&msg, NULL, 0, 0)) {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
   }

   UnregisterClass(THIS_CLASSNAME, hinstance);

   return (int) msg.wParam;
}
